# Autogenerated by nbdev

d = { 'settings': { 'branch': 'main',
                'doc_baseurl': '/lssm',
                'doc_host': 'https://franckalbinet.github.io',
                'git_url': 'https://github.com/franckalbinet/lssm',
                'lib_path': 'lssm'},
  'syms': { 'lssm.callbacks': { 'lssm.callbacks.BaseSchedCB': ('callbacks.html#baseschedcb', 'lssm/callbacks.py'),
                                'lssm.callbacks.BaseSchedCB.__init__': ('callbacks.html#baseschedcb.__init__', 'lssm/callbacks.py'),
                                'lssm.callbacks.BaseSchedCB._step': ('callbacks.html#baseschedcb._step', 'lssm/callbacks.py'),
                                'lssm.callbacks.BaseSchedCB.before_fit': ('callbacks.html#baseschedcb.before_fit', 'lssm/callbacks.py'),
                                'lssm.callbacks.BatchSchedCB': ('callbacks.html#batchschedcb', 'lssm/callbacks.py'),
                                'lssm.callbacks.BatchSchedCB.after_batch': ('callbacks.html#batchschedcb.after_batch', 'lssm/callbacks.py'),
                                'lssm.callbacks.BatchTransformCB': ('callbacks.html#batchtransformcb', 'lssm/callbacks.py'),
                                'lssm.callbacks.BatchTransformCB.__init__': ( 'callbacks.html#batchtransformcb.__init__',
                                                                              'lssm/callbacks.py'),
                                'lssm.callbacks.BatchTransformCB.before_batch': ( 'callbacks.html#batchtransformcb.before_batch',
                                                                                  'lssm/callbacks.py'),
                                'lssm.callbacks.Callback': ('callbacks.html#callback', 'lssm/callbacks.py'),
                                'lssm.callbacks.CancelBatchException': ('callbacks.html#cancelbatchexception', 'lssm/callbacks.py'),
                                'lssm.callbacks.CancelEpochException': ('callbacks.html#cancelepochexception', 'lssm/callbacks.py'),
                                'lssm.callbacks.CancelFitException': ('callbacks.html#cancelfitexception', 'lssm/callbacks.py'),
                                'lssm.callbacks.DeviceCB': ('callbacks.html#devicecb', 'lssm/callbacks.py'),
                                'lssm.callbacks.DeviceCB.__init__': ('callbacks.html#devicecb.__init__', 'lssm/callbacks.py'),
                                'lssm.callbacks.DeviceCB.before_batch': ('callbacks.html#devicecb.before_batch', 'lssm/callbacks.py'),
                                'lssm.callbacks.DeviceCB.before_fit': ('callbacks.html#devicecb.before_fit', 'lssm/callbacks.py'),
                                'lssm.callbacks.LRFinderCB': ('callbacks.html#lrfindercb', 'lssm/callbacks.py'),
                                'lssm.callbacks.LRFinderCB.__init__': ('callbacks.html#lrfindercb.__init__', 'lssm/callbacks.py'),
                                'lssm.callbacks.LRFinderCB.after_batch': ('callbacks.html#lrfindercb.after_batch', 'lssm/callbacks.py'),
                                'lssm.callbacks.LRFinderCB.before_fit': ('callbacks.html#lrfindercb.before_fit', 'lssm/callbacks.py'),
                                'lssm.callbacks.LRFinderCB.cleanup_fit': ('callbacks.html#lrfindercb.cleanup_fit', 'lssm/callbacks.py'),
                                'lssm.callbacks.MetricsCB': ('callbacks.html#metricscb', 'lssm/callbacks.py'),
                                'lssm.callbacks.MetricsCB.__init__': ('callbacks.html#metricscb.__init__', 'lssm/callbacks.py'),
                                'lssm.callbacks.MetricsCB._log': ('callbacks.html#metricscb._log', 'lssm/callbacks.py'),
                                'lssm.callbacks.MetricsCB.after_batch': ('callbacks.html#metricscb.after_batch', 'lssm/callbacks.py'),
                                'lssm.callbacks.MetricsCB.after_epoch': ('callbacks.html#metricscb.after_epoch', 'lssm/callbacks.py'),
                                'lssm.callbacks.MetricsCB.before_epoch': ('callbacks.html#metricscb.before_epoch', 'lssm/callbacks.py'),
                                'lssm.callbacks.MetricsCB.before_fit': ('callbacks.html#metricscb.before_fit', 'lssm/callbacks.py'),
                                'lssm.callbacks.ProgressCB': ('callbacks.html#progresscb', 'lssm/callbacks.py'),
                                'lssm.callbacks.ProgressCB.__init__': ('callbacks.html#progresscb.__init__', 'lssm/callbacks.py'),
                                'lssm.callbacks.ProgressCB._log': ('callbacks.html#progresscb._log', 'lssm/callbacks.py'),
                                'lssm.callbacks.ProgressCB.after_batch': ('callbacks.html#progresscb.after_batch', 'lssm/callbacks.py'),
                                'lssm.callbacks.ProgressCB.after_epoch': ('callbacks.html#progresscb.after_epoch', 'lssm/callbacks.py'),
                                'lssm.callbacks.ProgressCB.before_epoch': ('callbacks.html#progresscb.before_epoch', 'lssm/callbacks.py'),
                                'lssm.callbacks.ProgressCB.before_fit': ('callbacks.html#progresscb.before_fit', 'lssm/callbacks.py'),
                                'lssm.callbacks.SingleBatchCB': ('callbacks.html#singlebatchcb', 'lssm/callbacks.py'),
                                'lssm.callbacks.SingleBatchCB.after_batch': ( 'callbacks.html#singlebatchcb.after_batch',
                                                                              'lssm/callbacks.py'),
                                'lssm.callbacks.TrainCB': ('callbacks.html#traincb', 'lssm/callbacks.py'),
                                'lssm.callbacks.TrainCB.__init__': ('callbacks.html#traincb.__init__', 'lssm/callbacks.py'),
                                'lssm.callbacks.TrainCB.backward': ('callbacks.html#traincb.backward', 'lssm/callbacks.py'),
                                'lssm.callbacks.TrainCB.get_loss': ('callbacks.html#traincb.get_loss', 'lssm/callbacks.py'),
                                'lssm.callbacks.TrainCB.predict': ('callbacks.html#traincb.predict', 'lssm/callbacks.py'),
                                'lssm.callbacks.TrainCB.step': ('callbacks.html#traincb.step', 'lssm/callbacks.py'),
                                'lssm.callbacks.TrainCB.zero_grad': ('callbacks.html#traincb.zero_grad', 'lssm/callbacks.py'),
                                'lssm.callbacks.run_cbs': ('callbacks.html#run_cbs', 'lssm/callbacks.py'),
                                'lssm.callbacks.to_cpu': ('callbacks.html#to_cpu', 'lssm/callbacks.py'),
                                'lssm.callbacks.to_device': ('callbacks.html#to_device', 'lssm/callbacks.py')},
            'lssm.dataloaders': { 'lssm.dataloaders.CrossSpectraDataset': ('dataloaders.html#crossspectradataset', 'lssm/dataloaders.py'),
                                  'lssm.dataloaders.CrossSpectraDataset.__getitem__': ( 'dataloaders.html#crossspectradataset.__getitem__',
                                                                                        'lssm/dataloaders.py'),
                                  'lssm.dataloaders.CrossSpectraDataset.__init__': ( 'dataloaders.html#crossspectradataset.__init__',
                                                                                     'lssm/dataloaders.py'),
                                  'lssm.dataloaders.CrossSpectraDataset.__len__': ( 'dataloaders.html#crossspectradataset.__len__',
                                                                                    'lssm/dataloaders.py'),
                                  'lssm.dataloaders.SpectralDataset': ('dataloaders.html#spectraldataset', 'lssm/dataloaders.py'),
                                  'lssm.dataloaders.SpectralDataset.__getitem__': ( 'dataloaders.html#spectraldataset.__getitem__',
                                                                                    'lssm/dataloaders.py'),
                                  'lssm.dataloaders.SpectralDataset.__init__': ( 'dataloaders.html#spectraldataset.__init__',
                                                                                 'lssm/dataloaders.py'),
                                  'lssm.dataloaders.SpectralDataset.__len__': ( 'dataloaders.html#spectraldataset.__len__',
                                                                                'lssm/dataloaders.py'),
                                  'lssm.dataloaders.SpectralEmbeddingDataset': ( 'dataloaders.html#spectralembeddingdataset',
                                                                                 'lssm/dataloaders.py'),
                                  'lssm.dataloaders.SpectralEmbeddingDataset.__getitem__': ( 'dataloaders.html#spectralembeddingdataset.__getitem__',
                                                                                             'lssm/dataloaders.py'),
                                  'lssm.dataloaders.SpectralEmbeddingDataset.__init__': ( 'dataloaders.html#spectralembeddingdataset.__init__',
                                                                                          'lssm/dataloaders.py'),
                                  'lssm.dataloaders.SpectralEmbeddingDataset.__len__': ( 'dataloaders.html#spectralembeddingdataset.__len__',
                                                                                         'lssm/dataloaders.py'),
                                  'lssm.dataloaders.get_dls': ('dataloaders.html#get_dls', 'lssm/dataloaders.py')},
            'lssm.hooks': { 'lssm.hooks.Hook': ('hooks.html#hook', 'lssm/hooks.py'),
                            'lssm.hooks.Hook.__del__': ('hooks.html#hook.__del__', 'lssm/hooks.py'),
                            'lssm.hooks.Hook.__init__': ('hooks.html#hook.__init__', 'lssm/hooks.py'),
                            'lssm.hooks.Hook.remove': ('hooks.html#hook.remove', 'lssm/hooks.py'),
                            'lssm.hooks.ToyCNN': ('hooks.html#toycnn', 'lssm/hooks.py'),
                            'lssm.hooks.ToyCNN.__init__': ('hooks.html#toycnn.__init__', 'lssm/hooks.py'),
                            'lssm.hooks.ToyCNN.forward': ('hooks.html#toycnn.forward', 'lssm/hooks.py')},
            'lssm.learner': { 'lssm.learner.Learner': ('learner.html#learner', 'lssm/learner.py'),
                              'lssm.learner.Learner.__getattr__': ('learner.html#learner.__getattr__', 'lssm/learner.py'),
                              'lssm.learner.Learner.__init__': ('learner.html#learner.__init__', 'lssm/learner.py'),
                              'lssm.learner.Learner._fit': ('learner.html#learner._fit', 'lssm/learner.py'),
                              'lssm.learner.Learner._one_batch': ('learner.html#learner._one_batch', 'lssm/learner.py'),
                              'lssm.learner.Learner._one_epoch': ('learner.html#learner._one_epoch', 'lssm/learner.py'),
                              'lssm.learner.Learner.callback': ('learner.html#learner.callback', 'lssm/learner.py'),
                              'lssm.learner.Learner.fit': ('learner.html#learner.fit', 'lssm/learner.py'),
                              'lssm.learner.Learner.get_preds': ('learner.html#learner.get_preds', 'lssm/learner.py'),
                              'lssm.learner.Learner.one_epoch': ('learner.html#learner.one_epoch', 'lssm/learner.py'),
                              'lssm.learner.Learner.training': ('learner.html#learner.training', 'lssm/learner.py'),
                              'lssm.learner.lr_find': ('learner.html#lr_find', 'lssm/learner.py'),
                              'lssm.learner.with_cbs': ('learner.html#with_cbs', 'lssm/learner.py'),
                              'lssm.learner.with_cbs.__call__': ('learner.html#with_cbs.__call__', 'lssm/learner.py'),
                              'lssm.learner.with_cbs.__init__': ('learner.html#with_cbs.__init__', 'lssm/learner.py')},
            'lssm.loading': { 'lssm.loading.download': ('loading.html#download', 'lssm/loading.py'),
                              'lssm.loading.get_spectra_pair_idxs': ('loading.html#get_spectra_pair_idxs', 'lssm/loading.py'),
                              'lssm.loading.load_mir_kex_spike': ('loading.html#load_mir_kex_spike', 'lssm/loading.py'),
                              'lssm.loading.load_mir_ring_trial': ('loading.html#load_mir_ring_trial', 'lssm/loading.py'),
                              'lssm.loading.load_nir_kex_spike': ('loading.html#load_nir_kex_spike', 'lssm/loading.py'),
                              'lssm.loading.load_ossl': ('loading.html#load_ossl', 'lssm/loading.py')},
            'lssm.models': { 'lssm.models.CVAE': ('models.html#cvae', 'lssm/models.py'),
                             'lssm.models.CVAE.__init__': ('models.html#cvae.__init__', 'lssm/models.py'),
                             'lssm.models.CVAE.forward': ('models.html#cvae.forward', 'lssm/models.py'),
                             'lssm.models.CVAE.reparameterize': ('models.html#cvae.reparameterize', 'lssm/models.py'),
                             'lssm.models.Decoder': ('models.html#decoder', 'lssm/models.py'),
                             'lssm.models.Decoder.__init__': ('models.html#decoder.__init__', 'lssm/models.py'),
                             'lssm.models.Decoder.forward': ('models.html#decoder.forward', 'lssm/models.py'),
                             'lssm.models.Encoder': ('models.html#encoder', 'lssm/models.py'),
                             'lssm.models.Encoder.__init__': ('models.html#encoder.__init__', 'lssm/models.py'),
                             'lssm.models.Encoder.forward': ('models.html#encoder.forward', 'lssm/models.py'),
                             'lssm.models.GeneralRelu': ('models.html#generalrelu', 'lssm/models.py'),
                             'lssm.models.GeneralRelu.__init__': ('models.html#generalrelu.__init__', 'lssm/models.py'),
                             'lssm.models.GeneralRelu.forward': ('models.html#generalrelu.forward', 'lssm/models.py'),
                             'lssm.models.ResBlock': ('models.html#resblock', 'lssm/models.py'),
                             'lssm.models.ResBlock.__init__': ('models.html#resblock.__init__', 'lssm/models.py'),
                             'lssm.models.ResBlock.forward': ('models.html#resblock.forward', 'lssm/models.py'),
                             'lssm.models.ResBlockUp': ('models.html#resblockup', 'lssm/models.py'),
                             'lssm.models.ResBlockUp.__init__': ('models.html#resblockup.__init__', 'lssm/models.py'),
                             'lssm.models.ResBlockUp.forward': ('models.html#resblockup.forward', 'lssm/models.py'),
                             'lssm.models.conv': ('models.html#conv', 'lssm/models.py'),
                             'lssm.models.conv_block': ('models.html#conv_block', 'lssm/models.py'),
                             'lssm.models.init_weights': ('models.html#init_weights', 'lssm/models.py'),
                             'lssm.models.plot_func': ('models.html#plot_func', 'lssm/models.py')},
            'lssm.preprocessing': { 'lssm.preprocessing.BaselineALS': ('preprocessing.html#baselineals', 'lssm/preprocessing.py'),
                                    'lssm.preprocessing.BaselineALS.__init__': ( 'preprocessing.html#baselineals.__init__',
                                                                                 'lssm/preprocessing.py'),
                                    'lssm.preprocessing.BaselineALS._baseline_als': ( 'preprocessing.html#baselineals._baseline_als',
                                                                                      'lssm/preprocessing.py'),
                                    'lssm.preprocessing.BaselineALS.fit': ('preprocessing.html#baselineals.fit', 'lssm/preprocessing.py'),
                                    'lssm.preprocessing.BaselineALS.transform': ( 'preprocessing.html#baselineals.transform',
                                                                                  'lssm/preprocessing.py'),
                                    'lssm.preprocessing.ContinuumRemoval': ('preprocessing.html#continuumremoval', 'lssm/preprocessing.py'),
                                    'lssm.preprocessing.ContinuumRemoval.__init__': ( 'preprocessing.html#continuumremoval.__init__',
                                                                                      'lssm/preprocessing.py'),
                                    'lssm.preprocessing.ContinuumRemoval.fit': ( 'preprocessing.html#continuumremoval.fit',
                                                                                 'lssm/preprocessing.py'),
                                    'lssm.preprocessing.ContinuumRemoval.transform': ( 'preprocessing.html#continuumremoval.transform',
                                                                                       'lssm/preprocessing.py'),
                                    'lssm.preprocessing.Log1p': ('preprocessing.html#log1p', 'lssm/preprocessing.py'),
                                    'lssm.preprocessing.Log1p.fit': ('preprocessing.html#log1p.fit', 'lssm/preprocessing.py'),
                                    'lssm.preprocessing.Log1p.transform': ('preprocessing.html#log1p.transform', 'lssm/preprocessing.py'),
                                    'lssm.preprocessing.MeanCenter': ('preprocessing.html#meancenter', 'lssm/preprocessing.py'),
                                    'lssm.preprocessing.MeanCenter.fit': ('preprocessing.html#meancenter.fit', 'lssm/preprocessing.py'),
                                    'lssm.preprocessing.MeanCenter.transform': ( 'preprocessing.html#meancenter.transform',
                                                                                 'lssm/preprocessing.py'),
                                    'lssm.preprocessing.MinScaler': ('preprocessing.html#minscaler', 'lssm/preprocessing.py'),
                                    'lssm.preprocessing.MinScaler.fit': ('preprocessing.html#minscaler.fit', 'lssm/preprocessing.py'),
                                    'lssm.preprocessing.MinScaler.transform': ( 'preprocessing.html#minscaler.transform',
                                                                                'lssm/preprocessing.py'),
                                    'lssm.preprocessing.SNV': ('preprocessing.html#snv', 'lssm/preprocessing.py'),
                                    'lssm.preprocessing.SNV.fit': ('preprocessing.html#snv.fit', 'lssm/preprocessing.py'),
                                    'lssm.preprocessing.SNV.transform': ('preprocessing.html#snv.transform', 'lssm/preprocessing.py'),
                                    'lssm.preprocessing.SpikeDiff': ('preprocessing.html#spikediff', 'lssm/preprocessing.py'),
                                    'lssm.preprocessing.SpikeDiff.__init__': ( 'preprocessing.html#spikediff.__init__',
                                                                               'lssm/preprocessing.py'),
                                    'lssm.preprocessing.SpikeDiff._get_diffs_idx': ( 'preprocessing.html#spikediff._get_diffs_idx',
                                                                                     'lssm/preprocessing.py'),
                                    'lssm.preprocessing.SpikeDiff._get_levels_reps': ( 'preprocessing.html#spikediff._get_levels_reps',
                                                                                       'lssm/preprocessing.py'),
                                    'lssm.preprocessing.SpikeDiff.fit': ('preprocessing.html#spikediff.fit', 'lssm/preprocessing.py'),
                                    'lssm.preprocessing.SpikeDiff.transform': ( 'preprocessing.html#spikediff.transform',
                                                                                'lssm/preprocessing.py'),
                                    'lssm.preprocessing.SpikeMean': ('preprocessing.html#spikemean', 'lssm/preprocessing.py'),
                                    'lssm.preprocessing.SpikeMean.__init__': ( 'preprocessing.html#spikemean.__init__',
                                                                               'lssm/preprocessing.py'),
                                    'lssm.preprocessing.SpikeMean._get_smp_names': ( 'preprocessing.html#spikemean._get_smp_names',
                                                                                     'lssm/preprocessing.py'),
                                    'lssm.preprocessing.SpikeMean.fit': ('preprocessing.html#spikemean.fit', 'lssm/preprocessing.py'),
                                    'lssm.preprocessing.SpikeMean.transform': ( 'preprocessing.html#spikemean.transform',
                                                                                'lssm/preprocessing.py'),
                                    'lssm.preprocessing.TakeDerivative': ('preprocessing.html#takederivative', 'lssm/preprocessing.py'),
                                    'lssm.preprocessing.TakeDerivative.__init__': ( 'preprocessing.html#takederivative.__init__',
                                                                                    'lssm/preprocessing.py'),
                                    'lssm.preprocessing.TakeDerivative.fit': ( 'preprocessing.html#takederivative.fit',
                                                                               'lssm/preprocessing.py'),
                                    'lssm.preprocessing.TakeDerivative.transform': ( 'preprocessing.html#takederivative.transform',
                                                                                     'lssm/preprocessing.py'),
                                    'lssm.preprocessing.ToAbsorbance': ('preprocessing.html#toabsorbance', 'lssm/preprocessing.py'),
                                    'lssm.preprocessing.ToAbsorbance.__init__': ( 'preprocessing.html#toabsorbance.__init__',
                                                                                  'lssm/preprocessing.py'),
                                    'lssm.preprocessing.ToAbsorbance.fit': ('preprocessing.html#toabsorbance.fit', 'lssm/preprocessing.py'),
                                    'lssm.preprocessing.ToAbsorbance.transform': ( 'preprocessing.html#toabsorbance.transform',
                                                                                   'lssm/preprocessing.py')},
            'lssm.transforms': { 'lssm.transforms.GADFTfm': ('transforms.html#gadftfm', 'lssm/transforms.py'),
                                 'lssm.transforms.GADFTfm.__call__': ('transforms.html#gadftfm.__call__', 'lssm/transforms.py'),
                                 'lssm.transforms.GADFTfm.__init__': ('transforms.html#gadftfm.__init__', 'lssm/transforms.py'),
                                 'lssm.transforms.GADFTfm.rescale': ('transforms.html#gadftfm.rescale', 'lssm/transforms.py'),
                                 'lssm.transforms.SNVTfm': ('transforms.html#snvtfm', 'lssm/transforms.py'),
                                 'lssm.transforms.SNVTfm.__call__': ('transforms.html#snvtfm.__call__', 'lssm/transforms.py'),
                                 'lssm.transforms.SNVTfm.__init__': ('transforms.html#snvtfm.__init__', 'lssm/transforms.py'),
                                 'lssm.transforms.SNVTfm._apply_snv': ('transforms.html#snvtfm._apply_snv', 'lssm/transforms.py'),
                                 'lssm.transforms.StatsTfm': ('transforms.html#statstfm', 'lssm/transforms.py'),
                                 'lssm.transforms.StatsTfm.__call__': ('transforms.html#statstfm.__call__', 'lssm/transforms.py'),
                                 'lssm.transforms.StatsTfm.__init__': ('transforms.html#statstfm.__init__', 'lssm/transforms.py'),
                                 'lssm.transforms._resizeTfm': ('transforms.html#_resizetfm', 'lssm/transforms.py')},
            'lssm.visualization': {'lssm.visualization.plot_spectra': ('visualization.html#plot_spectra', 'lssm/visualization.py')}}}



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/07_models.ipynb.

# %% auto 0
__all__ = ['GeneralRelu', 'plot_func', 'init_weights', 'conv', 'conv_block', 'ResBlock', 'ResBlockUp', 'Encoder', 'Decoder',
           'CVAE']

# %% ../nbs/07_models.ipynb 3
from torch import optim, nn
from functools import partial

import matplotlib.pyplot as plt
import fastcore.all as fc

import torch
from torch.nn import init
import torchvision.transforms.functional as TF,torch.nn.functional as F

# %% ../nbs/07_models.ipynb 5
class GeneralRelu(nn.Module):
    "Relu unit as implemented in Fastai miniai package: https://github.com/fastai/minai"
    def __init__(self, leak=None, sub=None, maxv=None):
        super().__init__()
        self.leak, self.sub, self.maxv = leak, sub, maxv

    def forward(self, x):
        x = F.leaky_relu(x, self.leak) if self.leak is not None else F.relu(x)
        if self.sub is not None:
            x -= self.sub
        if self.maxv is not None:
            x.clamp_max_(self.maxv)
        return x

# %% ../nbs/07_models.ipynb 6
def plot_func(f, start=-5., end=5., steps=100, size=(5, 3)):
    "Plot a function (e.g activation function)."
    plt.figure(figsize=size)
    x = torch.linspace(start, end, steps)
    plt.plot(x, f(x))
    plt.grid(True, which='both', ls='--')
    plt.axhline(y=0, color='k', linewidth=0.7)
    plt.axvline(x=0, color='k', linewidth=0.7)

# %% ../nbs/07_models.ipynb 9
def init_weights(m, leaky=0.):
    "Initialize weights using kaiming normal method if Conv1d, ..., Linear."
    if isinstance(m, (nn.Conv1d, nn.Conv2d, nn.Conv3d, nn.Linear)):
        init.kaiming_normal_(m.weight, a=leaky)

# %% ../nbs/07_models.ipynb 11
def conv(
    ni: int, # Number of input channels
    nf: int, # Number of output channels 
    ks:int=3, # Kernel size
    stride:int=2, # Stride
    dim:int=1, # Convolution dimension (1 or 2)
    act:nn.Module=nn.ReLU, # Activation function
    norm:nn.Module=None, # Normalization layer type
    bias:bool=None, # Is bias needed?
    ):
    "Wrapping `torch.Conv1D` or `torch.Conv2D`."
    if bias is None:
        bias = not isinstance(
            norm, (nn.BatchNorm1d, nn.BatchNorm2d, nn.BatchNorm3d))
    _conv = nn.Conv1d if dim == 1 else nn.Conv2d
    layers = [_conv(ni, nf, stride=stride,
                        kernel_size=ks, padding=ks//2, bias=bias)]
    if norm:
        layers.append(norm(nf))
    if act:
        layers.append(act())
    return nn.Sequential(*layers)

# %% ../nbs/07_models.ipynb 15
def conv_block(
    ni: int, # Number of input channels
    nf: int, # Number of output channels 
    ks:int=3, # Kernel size
    stride:int=2, # Stride
    act:nn.Module=nn.ReLU, # Activation function
    norm:nn.Module=None, # Normalization layer type    
    ):
    "Initial Resnet conv block."
    return nn.Sequential(conv(ni, nf, stride=1, act=act, norm=norm, ks=ks),
                         conv(nf, nf, stride=stride, act=None, norm=norm, ks=ks))

# %% ../nbs/07_models.ipynb 18
class ResBlock(nn.Module):
    "1D Residual block."
    def __init__(self,
                 ni: int, # Number of input channels
                 nf: int, # Number of output channels 
                 ks:int=3, # Kernel size
                 stride:int=2, # Stride
                 act:nn.Module=nn.ReLU, # Activation function
                 norm:nn.Module=None, # Normalization layer type   
                 ):
        super().__init__()
        self.convs = conv_block(ni, nf, act=act, stride=stride, ks=ks, norm=norm)
        self.idconv = fc.noop if ni == nf else conv(
            ni, nf, ks=1, stride=1, act=None)
        self.pool = fc.noop if stride == 1 else nn.AvgPool1d(2, ceil_mode=True)
        self.act = act()

    def forward(self, x):
        return self.act(self.convs(x) + self.idconv(self.pool(x)))

# %% ../nbs/07_models.ipynb 21
class ResBlockUp(nn.Module):
    "1D Residual block including upsampling for a VAE decoder."
    def __init__(self, 
                 ni: int, # Number of input channels
                 nf: int, # Number of output channels 
                 ks:int=3, # Kernel size
                 stride:int=2, # Stride
                 act:nn.Module=nn.ReLU, # Activation function
                 norm:nn.Module=None, # Normalization layer type 
                 ):
        super().__init__()
        self.upsample = nn.Upsample(
            scale_factor=2, mode='linear', align_corners=True)
        self.resblock = ResBlock(ni, nf, act=act, norm=norm, stride=stride)

    def forward(self, x):
        return self.resblock(self.upsample(x))

# %% ../nbs/07_models.ipynb 24
class Encoder(nn.Module):
    "Convolutional VAE Encoder."
    def __init__(self, 
                 act:nn.Module=nn.ReLU, # Activation function
                 nfs:tuple=(8, 16, 32, 64, 128, 256), # Successive output channels size
                 norm:nn.Module=None, # Normalization layer type 
                 z_dim:int=2 # Latent space dimension
                 ):
        # super(Encoder, self).__init__()
        super().__init__()
        self.layers = [ResBlock(1, 8, stride=1, act=act, norm=norm)]
        self.layers += [ResBlock(nfs[i], nfs[i+1], act=act,
                                 norm=norm, stride=2) for i in range(len(nfs)-1)]
        self.layers += [nn.AdaptiveAvgPool1d(1), nn.Flatten()]
        self.resblocks = nn.Sequential(*self.layers)
        self.fc_mu = nn.Linear(nfs[-1], z_dim)
        self.fc_logvar = nn.Linear(nfs[-1], z_dim)

    def forward(self, x):
        x = self.resblocks(x)
        mu = self.fc_mu(x)
        logvar = self.fc_logvar(x)
        return mu, logvar

# %% ../nbs/07_models.ipynb 27
class Decoder(nn.Module):
    "Convolutional VAE Decoder."
    def __init__(self, 
                 in_shape:int, # Length of input (e.g a infrared spectra with 1676 wavenumbers)
                 act:nn.Module=nn.ReLU, # Activation function
                 nfs:tuple=(8, 16, 32, 64, 128, 256), # Successive output channels size
                 norm:nn.Module=None, # Normalization layer type 
                 z_dim:int=2 # Latent space dimension
                 ):
        # super(Decoder, self).__init__()
        super().__init__()
        self.feat_len = 1 + in_shape//2**(len(nfs)-1)
        self.lin = nn.Linear(z_dim, nfs[0]*self.feat_len, bias=False)
        self.nfs = nfs

        layers = []
        layers += [ResBlockUp(nfs[i], nfs[i+1], act=act,
                              norm=norm, stride=1) for i in range(len(nfs)-1)]
        layers.append(ResBlock(nfs[-1], 1, act=act,
                      norm=nn.BatchNorm1d, stride=1))
        self.resblocks = nn.Sequential(*layers)

    def forward(self, x):
        x = self.lin(x)
        x = x.reshape(-1, self.nfs[0], self.feat_len)
        return self.resblocks(x)

# %% ../nbs/07_models.ipynb 29
class CVAE(nn.Module):
    "1D Resnet-based convolutional variational autoencoder."
    def __init__(self, 
                 in_shape:int, # Length of input (e.g a infrared spectra with 1676 wavenumbers)
                 z_dim:int=2, # Latent space dimension
                 nfs:tuple=(8, 16, 32, 64, 128, 256), # Successive output channels size
                 ):
        # super(CVAE, self).__init__()
        super().__init__()
        self.in_shape = in_shape
        self.encoder = Encoder(nfs=nfs, z_dim=z_dim)
        self.decoder = Decoder(in_shape, nfs=nfs[::-1], z_dim=z_dim)

    def reparameterize(self, mu, logvar):
        std = torch.exp(0.5*logvar)
        eps = torch.randn_like(std)
        return mu + eps*std

    def forward(self, x):
        mu, logvar = self.encoder(x)
        z = self.reparameterize(mu, logvar)
        x_rec = self.decoder(z)
        strip_idx = x_rec.shape[-1] - self.in_shape
        return x_rec[:, :, :-strip_idx], mu, logvar



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/01_visualization.ipynb.

# %% auto 0
__all__ = ['plot_spectra']

# %% ../nbs/01_visualization.ipynb 3
from pathlib import Path

import numpy as np

import seaborn as sns
import matplotlib.pyplot as plt

# %% ../nbs/01_visualization.ipynb 5
def plot_spectra(X, X_names, sample=50, 
                 ascending=True,
                 alpha=0.8, color='#333', 
                 xlabel='Wavenumber',
                 ylabel='Absorbance',
                 title=None,
                 figsize=(20, 4)):
    sns.set_style("whitegrid")
    sns.despine()
    
    plt.figure(figsize=figsize)
    
    if len(X.shape) > 1:
        idx = np.random.randint(X.shape[0], size=sample)
        X = X[idx,:]
    else:
        X = X.reshape(1, -1)
        
    _min, _max = np.min(X_names), np.max(X_names)
    _order = [_min, _max] if ascending else [_max, _min]
    plt.xlim(*_order)
    
    for i in range(X.shape[0]): 
        sns.lineplot(x=X_names, y=X[i,:], lw=1, c=color, alpha=alpha)
    
    plt.locator_params(axis="x", nbins=20)   
    plt.title(title) 
    plt.xlabel(xlabel)
    plt.ylabel(ylabel);
    



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/03_dataloaders.ipynb.

# %% auto 0
__all__ = ['get_dls', 'CrossSpectraDataset', 'SpectralDataset', 'SpectralEmbeddingDataset']

# %% ../nbs/03_dataloaders.ipynb 2
from pathlib import Path
from tqdm import tqdm
from collections import namedtuple
import fastcore.all as fc

import numpy as np
import pandas as pd
import torch
from torch.utils.data import Dataset, DataLoader
from torch import tensor

# %% ../nbs/03_dataloaders.ipynb 5
def get_dls(
    train_ds:Dataset, # Train dataset
    valid_ds:Dataset, # Valid dataset
    bs:int, # Batch size 
    **kwargs
    ) -> DataLoader:  # NamedTuple `DataLoader`with `.train` and `.valid` field names
    "Train and valid dataloaders."
    Dataloaders = namedtuple('Dataloader', ['train', 'valid'])
    return Dataloaders(
        DataLoader(train_ds, batch_size=bs, shuffle=True, **kwargs),
        DataLoader(valid_ds, batch_size=bs*2, **kwargs))


# %% ../nbs/03_dataloaders.ipynb 6
class CrossSpectraDataset(Dataset):
    "Custom Pytorch dataset for IR instrument cross-calibration."
    def __init__(self, 
                 df:pd.DataFrame, # Infrared spectra with associated sample id
                 pair_idxs:list, # List of tuples (pairs) of replicate indices to sample from the dataframe 
                 spectra_at:int=2 # Index of the column where the spectroscopy data starts
                 ):
        self.data = tensor(df.iloc[:, spectra_at:].to_numpy(),
                           dtype=torch.float32)
        self.pair_idxs = [list(p) for p in pair_idxs]

    def __len__(self):
        return len(self.pair_idxs)

    def __getitem__(self, idx):
        X, y = self.data[self.pair_idxs[idx]]
        return X[None, :], y[None, :]

# %% ../nbs/03_dataloaders.ipynb 11
class SpectralDataset(Dataset):
    "(Infrared Spectra, soil property) custom PyTorch Dataset."

    def __init__(self,
                 X:np.ndarray, # Spectra
                 y:np.ndarray, # Analyte
                 metadata:np.ndarray=None
                 ):
        self.X = X
        self.y = y
        self.metadata = metadata

    def __len__(self):
        return len(self.X)

    def __getitem__(self, idx):
        return (tensor(self.X[idx, :][None, :], dtype=torch.float32),
                tensor(self.y[idx, :], dtype=torch.float32))

# %% ../nbs/03_dataloaders.ipynb 14
class SpectralEmbeddingDataset(Dataset):
    "(Infrared Spectra, Infrared Spectra) custom PyTorch Dataset."

    def __init__(self,
                 X:np.ndarray, # Spectra
                 ):
        self.X = X

    def __len__(self):
        return len(self.X)

    def __getitem__(self, idx):
        return (tensor(self.X[idx, :][None, :], dtype=torch.float32),
                tensor(self.X[idx, :][None, :], dtype=torch.float32))



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/00_loading.ipynb.

# %% auto 0
__all__ = ['PATH_OSSL_ALL_L0_V1_2', 'PATH_OSSL_ALL_L1_V1_2', 'CFGS', 'download', 'load_ossl', 'load_mir_ring_trial',
           'get_spectra_pair_idxs', 'load_mir_kex_spike', 'load_nir_kex_spike']

# %% ../nbs/00_loading.ipynb 3
from pathlib import Path
from tqdm import tqdm
from typing import Union, List
import re
import itertools
import fastdownload as fd
import fastcore.all as fc

import pandas as pd
import numpy as np
from sklearn.preprocessing import LabelEncoder

import warnings
warnings.filterwarnings('ignore')

# %% ../nbs/00_loading.ipynb 6
PATH_OSSL_ALL_L0_V1_2 = 'https://storage.googleapis.com/soilspec4gg-public/ossl_all_L0_v1.2.csv.gz'
PATH_OSSL_ALL_L1_V1_2 = 'https://storage.googleapis.com/soilspec4gg-public/ossl_all_L0_v1.2.csv.gz'

CFGS = {
    'visnir': {'ref_col': 'scan_visnir.1500_ref', 'range': [400, 2500]},
    'mir': {'ref_col': 'scan_mir.1500_abs', 'range': [600, 4000]}
}

# %% ../nbs/00_loading.ipynb 7
def download(url: str,  # url to dowload data from
             dest_dir: str,  # directory to download data to
             ) -> None:
    "Download data available at `url` into the `dest` directory (creates it on the way if does not exist)."
    if not dest_dir.exists():
        fc.mkdir(dest_dir, parents=True)
    return fd.download_url(url, dest_dir)

# %% ../nbs/00_loading.ipynb 8
def load_ossl(analytes: Union[str, List[str]],  # Using OSSL's analytes naming conventions
              spectra_type: str = 'visnir',  # Possible values: 'mir', 'visnir'
              src: Path = Path.home() / '.lssm/data/ossl',  # directory containing the data
              debug:bool=False # return unprocessed loaded data directly for further investigation
              ):
    "Load all available OSSL data and filter it by spectra type and analytes of interest"

    url = PATH_OSSL_ALL_L1_V1_2
    fname = src / Path(PATH_OSSL_ALL_L1_V1_2).name
    if not fname.exists():
        print('Downloading & saving to: ', str(fname))
        download(url, src)

    print('Reading & selecting data ...')

    df = pd.read_csv(fname, compression='infer', low_memory=True)
    
    if debug: return df
    
    analytes = [analytes] if isinstance(analytes, str) else analytes

    subset = analytes + [CFGS[spectra_type]['ref_col']]
    df = df.dropna(subset=subset, how='any')

    cols_ref = [name for name in df.columns if f'scan_{spectra_type}.' in name]
    X = df[cols_ref].values

    y = df[analytes].values
    smp_idx = df['id.layer_uuid_txt'].values

    ds_name_encoder = LabelEncoder()
    ds_name = ds_name_encoder.fit_transform(df['dataset.code_ascii_txt'])

    pattern = r"scan_{}\.(\d+)_".format(spectra_type)
    X_names = np.array([int(re.search(pattern, name).group(1)) for name in df.columns
                        if re.search(pattern, name)])

    lower_limit, upper_limit = CFGS[spectra_type]['range']
    idxs = np.where((X_names >= lower_limit) & (X_names <= upper_limit))[0]

    return X[:, idxs], y, X_names[idxs], smp_idx, ds_name, ds_name_encoder.classes_


# %% ../nbs/00_loading.ipynb 15
def load_mir_ring_trial(fname): return pd.read_csv(fname)

# %% ../nbs/00_loading.ipynb 18
def get_spectra_pair_idxs(df):
    """
    Retrieve index pairs of replicated spectra, representing measurements taken from the same soil sample
    but using different instruments.
    """
    pair_idxs = []
    for smp_id in df.sample_id.unique():
        indices = df[df.sample_id == smp_id].index
        pair_idxs.extend(list(itertools.product(indices, repeat=2)))
    return pair_idxs


# %% ../nbs/00_loading.ipynb 23
def load_mir_kex_spike(src_dir):
    """
    Load MIR spectra of K spiked soil samples.
    
    Parameters:
    src_dir (Path-like object): Directory containing the spectra files.
    
    Returns:
    tuple: Tuple containing the array of absorbance values, 
           array of wavenumbers (columns), and array of sample names (rows).
    """
    pattern = r'-\d-\d$'
    fnames = [f for f in src_dir.ls() if re.search(pattern, f.stem)]
    
    dfs = [pd.read_csv(fname, header=None, names=['wavenumber', 'absorbance'])
           .query('649 < wavenumber < 4000')
           .assign(name=fname.stem) for fname in fnames]
    
    df_combined = pd.concat(dfs).pivot_table(values='absorbance', index='name', columns='wavenumber')
    
    return df_combined.values, df_combined.columns.values, df_combined.index.values.astype('U')


# %% ../nbs/00_loading.ipynb 28
def load_nir_kex_spike(fname):
    """
    Load NIR spectra of K spiked soil samples.
    
    Parameters:
    fname (str or Path-like object): File name or path of the Excel file.
    
    Returns:
    tuple: Tuple containing the array of spectral values, 
           array of wavenumbers (columns), and array of sample names (rows).
    """
    df = pd.read_excel(fname, sheet_name='Results', index_col='Sample ID')
    df.index.name = 'name'
    df.columns.name = 'wavenumber'
    
    return df.values, df.columns.values, df.index.values.astype('U')



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/02_preprocessing.ipynb.

# %% auto 0
__all__ = ['ToAbsorbance', 'ContinuumRemoval', 'SNV', 'Log1p', 'SpikeDiff', 'TakeDerivative', 'MinScaler', 'MeanCenter',
           'BaselineALS', 'SpikeMean']

# %% ../nbs/02_preprocessing.ipynb 3
from pathlib import Path
from tqdm import tqdm
import re

import fastcore.all as fc
from fastcore.basics import patch

import pandas as pd
import numpy as np

from sklearn.base import BaseEstimator, TransformerMixin
from scipy.spatial import ConvexHull
from scipy.interpolate import interp1d
from scipy.signal import savgol_filter
from scipy.sparse import diags
from scipy.sparse.linalg import spsolve

# %% ../nbs/02_preprocessing.ipynb 6
class ToAbsorbance(BaseEstimator, TransformerMixin):
    """Transform Reflectance to Absorbance"""
    def __init__(self, eps=1e-5): self.eps = eps
    def fit(self, X, y=None): return self
    def transform(self, X, y=None): 
        X[X < 0] = 0
        return -np.log10(X + self.eps)

# %% ../nbs/02_preprocessing.ipynb 10
class ContinuumRemoval(BaseEstimator, TransformerMixin):
    """Creates continnum removal custom transformer"""

    def __init__(self, wls):
        fc.store_attr()

    def fit(self, X, y=None):
        return self

    def transform(self, X, y=None):
        continuum_removed_spectra = np.zeros_like(X)
        
        for i, spectrum in enumerate(tqdm(X)):
            points = np.c_[self.wls, spectrum]
            x, y = points.T 
            # trick to exclude lower part of convex hull 
            augmented = np.concatenate([points, [(x[0], np.max(y)+1), (x[-1], np.max(y)+1)]], axis=0)
            hull = ConvexHull(augmented)
            continuum_points = points[np.sort([v for v in hull.vertices if v < len(points)])]
            continuum_function = interp1d(*continuum_points.T)
            continuum_removed_spectra[i, :] = y / continuum_function(x)
    
        return continuum_removed_spectra

# %% ../nbs/02_preprocessing.ipynb 13
class SNV(BaseEstimator, TransformerMixin):
    """Creates scikit-learn SNV custom transformer"""
    def fit(self, X, y=None): return self
    def transform(self, X, y=None):
        mean, std = np.mean(X, axis=1).reshape(-1, 1), np.std(X, axis=1).reshape(-1, 1)
        return (X - mean)/std

# %% ../nbs/02_preprocessing.ipynb 15
class Log1p(BaseEstimator, TransformerMixin):
    """Creates scikit-learn `np.log1p` target custom transformer"""
    def fit(self, y, X=None): return self
    def transform(self, y, X=None): return np.log1p(y)

# %% ../nbs/02_preprocessing.ipynb 16
class SpikeDiff(BaseEstimator, TransformerMixin):
    """
    Scikit-learn transformer for taking the difference of spiked sample spectra.

    Attributes:
    names (list): List of names to be processed.
    idx (numpy.ndarray): Indices for differences in spectra.
    """
    def __init__(self, names):
        fc.store_attr()
        self.idx = self._get_diffs_idx(self._get_levels_reps())
        
    def _get_levels_reps(self):
        pattern = r'-(\d+)-(\d+)'
        level_rep = [re.search(pattern, name).groups() for name in self.names]

        data = {'index': [], 'level': [], 'replicate': []}
        for i, (level, replicate) in enumerate(level_rep):
            data['index'].append(i)
            data['level'].append(int(level))
            data['replicate'].append(int(replicate))
        
        return pd.DataFrame(data)
    
    def _get_diffs_idx(self, df):
        df_cross = df.merge(df, how='cross')
        return df_cross[df_cross.level_x > df_cross.level_y][['index_x', 'index_y']].values
    
    def fit(self, X, y=None): return self
    
    def transform(self, X):
        return X[self.idx,:][:, 1, :] - X[self.idx,:][:, 0, :]

# %% ../nbs/02_preprocessing.ipynb 19
# SpikeDiff(names[mask_smp]).fit_transform(X[mask_smp,:])

# %% ../nbs/02_preprocessing.ipynb 20
class TakeDerivative(BaseEstimator, TransformerMixin):
    """Creates scikit-learn derivation custom transformer

    Args:
        window_length: int, optional
            Specify savgol filter smoothing window length

        polyorder: int, optional
            Specify order of the polynom used to interpolate derived signal

        deriv: int, optional
            Specify derivation degree

    Returns:
        scikit-learn custom transformer
    """
    def __init__(self, window_length=11, polyorder=1, deriv=1):
        self.window_length = window_length
        self.polyorder = polyorder
        self.deriv = deriv

    def fit(self, X, y=None):
        return self

    def transform(self, X, y=None):
        return savgol_filter(X, self.window_length, self.polyorder, self.deriv)
     

# %% ../nbs/02_preprocessing.ipynb 23
class MinScaler(BaseEstimator, TransformerMixin):
    def fit(self, X, y=None): return self
    def transform(self, X, y=None):     
        return X - X.min(axis=1, keepdims=True)    

# %% ../nbs/02_preprocessing.ipynb 25
class MeanCenter(BaseEstimator, TransformerMixin):
    def fit(self, X, y=None): return self
    def transform(self, X, y=None):     
        return X - X.mean(axis=1, keepdims=True)    

# %% ../nbs/02_preprocessing.ipynb 28
class BaselineALS(BaseEstimator, TransformerMixin):
    def __init__(self, lam = 1e5, p = 0.01, niter=10):
        fc.store_attr()

    def _baseline_als(self, x):
        L = len(x)
        D = diags([1,-2,1], [0,-1,-2], shape=(L,L-2))
        w = np.ones(L)
        for i in range(self.niter):
            W = diags([w], [0], shape=(L,L))
            Z = W + self.lam * D.dot(D.transpose())
            z = spsolve(Z, w*x)
            w = self.p * (x > z) + (1-self.p) * (x < z)
        return z
    
    def fit(self, X, y=None):
        return self

    def transform(self, X, y=None):
        corrected_spectra = np.zeros_like(X)
        for i, spectrum in enumerate(tqdm(X)):
            baseline = self._baseline_als(spectrum)
            corrected_spectra[i, :] = spectrum - baseline
    
        return corrected_spectra

# %% ../nbs/02_preprocessing.ipynb 31
class SpikeMean(BaseEstimator, TransformerMixin):
    def __init__(self, names):
        fc.store_attr()
        self.smp_names = self._get_smp_names()
    
    def _get_smp_names(self):
        return np.unique(np.array([re.sub(r"-\d$", "", name) for name in self.names]))
        
    def fit(self, X, y=None): return self
    
    def transform(self, X):
        means = []
        for name in self.smp_names:
            mask_smp = np.char.find(self.names, name) == 0
            means.append(X[mask_smp].mean(axis=0))
            
        return np.array(means), self.smp_names



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/04_transforms.ipynb.

# %% auto 0
__all__ = ['GADFTfm', 'StatsTfm', 'SNVTfm']

# %% ../nbs/04_transforms.ipynb 3
from pathlib import Path

import numpy as np
import fastcore.all as fc

import torch
from torchvision import transforms as T


# %% ../nbs/04_transforms.ipynb 5
class GADFTfm(): 
    """
    Transform batch of spectra S (B, 1, len(S)) into their Grammian Difference Matrix Field (GADF) of shape (B, 1, H, W)
    
    Notes:
    https://arxiv.org/pdf/1506.00327.pdf
    """
    def __init__(self, neg=True):
        self.neg = neg
        
    def rescale(self, x):
        m = torch.min(x, dim=-1, keepdim=True).values
        M = torch.max(x, dim=-1, keepdim=True).values
        return ((x - M) + (x - m)) / (M - m) if self.neg else (x - m) / (M - m)

    def __call__(self, 
                 b # Batch of spectra S: (B, 1, len(S))
                 ): 
        x, y, *metadata = b
        X, I = self.rescale(x), torch.ones_like(x)
        K = torch.sqrt(I - torch.square(X))
        x = (torch.matmul(torch.transpose(K, 1, 2), X) - 
             torch.matmul(torch.transpose(X, 1, 2), K)).unsqueeze_(1)
        return x, y, *metadata

# %% ../nbs/04_transforms.ipynb 9
def _resizeTfm(b, size=224):
    "Resize image"
    x, y, *metadata  = b
    return T.Resize(size)(x), y, *metadata

# %% ../nbs/04_transforms.ipynb 11
class StatsTfm(): 
    """
    Set pre-trained statistics.
    """
    def __init__(self, 
                 cfgs
                 ):
        self.mean = np.array(cfgs['mean']).mean()
        self.std = np.array(cfgs['std']).mean()

    def __call__(self, b): 
        x, y, *metadata = b
        means = torch.mean(x, (2, 3), keepdim=True)
        stds = torch.std(x, (2, 3), keepdim=True)
        x = (((x - means) / stds) * self.std) + self.mean
        return x, y, *metadata

# %% ../nbs/04_transforms.ipynb 12
class SNVTfm(): 
    "Apply SNV to input or to both input and output when used with a Variational Auto-Encoder."
    def __init__(self, 
                 is_VAE:bool=False # If VAE apply SNV on both input and output.
                 ):
        self.is_VAE = is_VAE

    def _apply_snv(self, data):
        mean = torch.mean(data, axis=-1, keepdim=True)
        std = torch.std(data, axis=-1, keepdim=True)
        return (data - mean) / std
    
    def __call__(self, b): 
        if fc.isinstance_str(b, 'tuple'):
            x, y = b
            if self.is_VAE:
                b = torch.concat(b)
                b = self._apply_snv(b)
                bs = len(x)
                return b[:bs], b[bs:]
            else:
                x = self._apply_snv(x)
                return x, y
        else:
            return self._apply_snv(b)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/08_hooks.ipynb.

# %% auto 0
__all__ = ['ToyCNN', 'Hook']

# %% ../nbs/08_hooks.ipynb 3
# from copy import copy
# import math
# import fastcore.all as fc
# from fastprogress import progress_bar, master_bar
# from operator import attrgetter
# from collections.abc import Mapping

# import matplotlib.pyplot as plt

# import torch
# from torchvision import transforms as T
# from torcheval.metrics import Mean

# %% ../nbs/08_hooks.ipynb 4
from sklearn.pipeline import Pipeline
from sklearn.model_selection import train_test_split

from .loading import load_ossl
from .models import conv
from .callbacks import to_cpu
from .preprocessing import Log1p, SNV
from .dataloaders import SpectralDataset, get_dls

import torch
from torch import nn, tensor

# %% ../nbs/08_hooks.ipynb 7
from functools import partial
from torch.optim import lr_scheduler

from torcheval.metrics import R2Score
from .callbacks import (MetricsCB, BatchSchedCB, BatchTransformCB,
                            DeviceCB, TrainCB, ProgressCB)
from .transforms import GADFTfm, _resizeTfm, StatsTfm
from .learner import Learner

# %% ../nbs/08_hooks.ipynb 8
class ToyCNN(nn.Module):
    def __init__(self):
        super().__init__()
        self.layers = nn.Sequential(
            conv(1, 8, ks=5),        # 8, 851
            conv(8, 16),             # 16, 426
            conv(16, 32),            # 32, 213
            conv(32, 64),            # 64, 107
            nn.Flatten(),
            nn.Linear(64*107, 1))
        
    def forward(self, x):
        return self.layers(x)

# %% ../nbs/08_hooks.ipynb 11
class Hook():
    def __init__(self, m, f): self.hook = m.register_forward_hook(partial(f, self))
    def remove(self): self.hook.remove()
    def __del__(self): self.remove()
